VERSION 1.0 CLASS
BEGIN
  MultiUse = -1  'True
  Persistable = 0  'NotPersistable
  DataBindingBehavior = 0  'vbNone
  DataSourceBehavior  = 0  'vbNone
  MTSTransactionMode  = 0  'NotAnMTSObject
END
Attribute VB_Name = "AI8"
Attribute VB_GlobalNameSpace = False
Attribute VB_Creatable = True
Attribute VB_PredeclaredId = False
Attribute VB_Exposed = True
'##CD Simple PathFinding and AI class. The path finding use a node graph system (like Half-Life) : you have to add nodes, create the graphs and then you can find paths.

'

' AI Class
'----------
' by Sylvain

' Path finding with nodes

' 10-08-00 : Creation

Private Type AINodes
   NodePosition As D3DVECTOR
   NodeAround() As Integer
   NodeAroundNum As Integer
   Flags As Integer
   NodeDistance() As Single
End Type

Private Type AIGraph
 Node() As AINodes
 NumNode As Integer
End Type

Private ResearchRadius, MaxChangeAltitude
Private Graph As AIGraph

Public Sub AddNode(Position As D3DVECTOR)
  '##BD Add a node to the path finding system nodes list. <i><b>Note :</b> When all the nodes are added, you have to call the AICreateGraph function.</i>
  '##PD Position Position in the 3d world of the new node.
  Graph.NumNode = Graph.NumNode + 1
  ReDim Preserve Graph.Node(Graph.NumNode)
  Graph.Node(Graph.NumNode).NodePosition = Position
End Sub


Public Sub CreateAIGraph()
  
  '##BD Links the nodes each others in order to speed up the path finding. <b><i>Note :</b> You must call this method before using any path finding method. </i>
  
  On Error Resume Next
  Dim i As Integer, Node As Integer, Distance As Single
  
  AddLog "AI Graph creating... " + CStr(Graph.NumNode) + " nodes"
  If DisplayConsole = True Then
    LocalConsole.WriteConsole "AI Graph creating...", RGB(0, 256, 0)
    FlipToScreen
  End If
  '
  ' Create links between the AI nodes.
  '
  OldCollisionDist = PrecisionCollide
  For i = 1 To Graph.NumNode
      ReDim Graph.Node(i).NodeDistance(0)
      ReDim Graph.Node(i).NodeAround(0)
      For Node = 1 To Graph.NumNode
       If Node <> i Then
        Distance = VModulus(VSubtract(Graph.Node(i).NodePosition, Graph.Node(Node).NodePosition))
        PrecisionCollide = Distance
        If Distance < ResearchRadius Then
          If LocalScene.Collide(Graph.Node(i).NodePosition, Graph.Node(Node).NodePosition) = False And (Graph.Node(i).NodePosition.y - Graph.Node(Node).NodePosition.y) < MaxChangeAltitude Then
            With Graph.Node(i)
             ReDim Preserve .NodeAround(UBound(.NodeAround) + 1)
             ReDim Preserve .NodeDistance(UBound(.NodeDistance) + 1)
             .NodeAroundNum = .NodeAroundNum + 1
             .NodeAround(UBound(.NodeAround)) = Node
             .NodeDistance(UBound(.NodeDistance)) = Distance
             
            End With
          End If
        End If
       End If
      Next Node
  Next i
  'Let a place to the end and start point
  ReDim Preserve Graph.Node(UBound(Graph.Node) + 2)
  PrecisionCollide = OldCollisionDist
End Sub

Public Sub ExportGraph(Filename As String)
  '##BD Export the current AI graph to a file
  '##PD filename Filename to save
  Open Filename For Binary As 7
  Put #7, , "TVAI"
  Put #7, , Graph
  Close #7
End Sub

Public Function FindPath(StartPos As D3DVECTOR, EndPos As D3DVECTOR, ByRef StorePath As Path8) As Boolean
    '##BD Find a path from the start vector to the end vector. Returns True if a path has been found, False else.
    '##PD StartPos Start vector in the 3d world
    '##PD EndPos End vector in the 3d world
    '##PD Path Path in what the found path will be stored.
    On Error GoTo treaterr
    Dim StartNode As Integer
    Dim EndNode As Integer
    Dim Score() As Single
    Dim BestScore As Single
    Dim i As Integer, N As Integer
    Dim NodePath() As Integer
    
    ReDim Score(Graph.NumNode + 2)
    ReDim NodePath(0)
    Graph.Node(Graph.NumNode + 1).NodePosition = StartPos
    Graph.Node(Graph.NumNode + 2).NodePosition = EndPos
    For i = Graph.NumNode + 1 To Graph.NumNode + 2
      ReDim Graph.Node(i).NodeDistance(0)
      ReDim Graph.Node(i).NodeAround(0)
      For Node = 1 To Graph.NumNode
       If Node <> i Then
        Distance = VModulus(VSubtract(Graph.Node(i).NodePosition, Graph.Node(Node).NodePosition))
        PrecisionCollide = Distance
        If Distance < ResearchRadius Then
          If LocalScene.Collide(Graph.Node(i).NodePosition, Graph.Node(Node).NodePosition) = False Then
            With Graph.Node(i)
             ReDim Preserve .NodeAround(UBound(.NodeAround) + 1)
             ReDim Preserve .NodeDistance(UBound(.NodeDistance) + 1)
             ReDim Preserve Graph.Node(Node).NodeAround(Graph.Node(Node).NodeAroundNum + 1)
             ReDim Preserve Graph.Node(Node).NodeDistance(Graph.Node(Node).NodeAroundNum + 1)
             
             Graph.Node(Node).NodeAround(UBound(Graph.Node(Node).NodeAround)) = i
             Graph.Node(Node).NodeDistance(UBound(Graph.Node(Node).NodeDistance)) = Distance
             .NodeAround(UBound(.NodeAround)) = Node
             .NodeDistance(UBound(.NodeDistance)) = Distance
            End With
          End If
        End If
       End If
      Next Node
  Next i
    
    ' Voodoo VB Path finding algorithm.
    Score(Graph.NumNode + 2) = 1
    
   'Create a node map from the end node
   errsrt = "Map String"
   Do
    Changed = False
    For i = 1 To Graph.NumNode + 2
      With Graph.Node(i)
        If Score(i) <> 0 Then
          For N = 1 To UBound(.NodeAround)
            If Score(i) + .NodeDistance(N) < Score(.NodeAround(N)) Or Score(.NodeAround(N)) = 0 Then
               Score(.NodeAround(N)) = Score(i) + .NodeDistance(N)
               Changed = True
            End If
          Next N
        End If
      End With
    Next i
   Loop Until Changed = False

   'Now search the path
   S = Graph.NumNode + 1
   Do
    BestNode = -1
    BestScore = 10000000  'let's take a impossible number
    For i = 1 To UBound(Graph.Node(S).NodeAround)
      Index = Graph.Node(S).NodeAround(i)
      If Score(Index) < BestScore And Score(Index) <> 0 Then
         BestScore = Score(Index)
         BestNode = Index
      End If
    Next i
    DoEvents
    If BestNode = -1 Then
      'path no found
      AddLog "No found in searching"
      FindPath = False
      Exit Function
    End If
    ReDim Preserve NodePath(UBound(NodePath) + 1)
    NodePath(UBound(NodePath)) = BestNode
    If BestNode = Graph.NumNode + 2 Then
     Exit Do
    End If
    S = BestNode
   Loop
  Set StorePath = New Path8
  StorePath.AddPathNode Graph.Node(Graph.NumNode + 1).NodePosition
  For i = 1 To UBound(NodePath)
    StorePath.AddPathNode Graph.Node(NodePath(i)).NodePosition
  Next i
  StorePath.AddPathNode Graph.Node(Graph.NumNode + 2).NodePosition
  FindPath = True
   
  Exit Function
treaterr:
  AddLog "Error in FindPath " + CStr(Err.Description) + " " + CStr(Err.Number)
  Resume Next
End Function



Sub ImportFromBSP(BSP As BSPTree8)
  '##BD Imports the AI nodes from Half-Life info_nodes
   Dim Ent As String
   Ent = BSP.GetEntity
  Dim InfoNode As String
  Dim TempToken As String, TempLeft As String
  Dim TokEnd As Long, TokStart As Long
  Dim a As D3DVECTOR
  TempLeft = Ent
  If DisplayConsole = True Then
        '  LocalConsole.WriteConsole "BSP converter : Info_node -> AI Graph ", RGB(256, 0, 256)
        '  FlipToScreen
  End If
 
  Do
    TokStart = InStr(TempLeft, "{")
    TokEnd = InStr(TempLeft, "}")
    TempToken = Mid$(TempLeft, TokStart, TokEnd - TokStart)
    TempLeft = Mid$(TempLeft, TokEnd + 1)
    If InStr(TempToken, "info_node") > 0 Then
        InfoNode = BSP.GetKeyValue(TempToken, "origin")
        a = BSP.TransformBSPVector(BSP.TransformKeyInVector(InfoNode))
        AddNode a
        If DisplayConsole = True Then
         ' LocalConsole.WriteConsole "Add info_node :" + CStr(a.X) + " " + CStr(a.y) + " " + CStr(a.z), RGB(0, 0, 256)
         ' FlipToScreen
        End If
    End If
  Loop Until Len(TempLeft) < 10
  CreateAIGraph
End Sub

Public Sub ImportGraph(Filename As String)
  '##BD Import a AI graph from a file.
  '##PD Filename AI graph file
  TestPAKFile Filename
  Open Filename For Binary As 7
    Dim magic As String * 4
    Get #7, , magic
    If magic <> "TVAI" Then AddLog "AI Import graph error : " + CStr(Filename) + " isn't a good TrueVision AI Graph file": Exit Sub
    Get #7, , Graph
  Close 7
  DeletePAKTemp Filename
End Sub
Private Function SearchFirstNode(Pos As D3DVECTOR) As Integer
  Dim BestNode As Integer, BestDistance As Single, i As Integer, Distance As Single
  BestDistance = 1000000
  'Search the best nearest node for the vector
  BestNode = -1
  For i = 1 To Graph.NumNode
    Distance = VModulus(VSubtract(Pos, Graph.Node(i).NodePosition))
    If Distance < ResearchRadius Then
      PrecisionCollide = Distance
      If LocalScene.Collide(Pos, Graph.Node(i).NodePosition) = False Then
        If Distance < BestDistance Then
          BestNode = i
          BestDistance = Distance
        End If
      End If
    End If
  Next i
  
  SearchFirstNode = BestNode
End Function

Sub SetFindPathParameters(ResearchRadius, MaxChangeAltitude)
  '##BD Set Pathfinding parameters.
  '##PD ResearchRadius The player can 'see' the node when the distance between the node and the player is less than this number
  '##PD MaxChangeAltitude Decides at what altitude the player can walk (from a node to another node)
  ResearchRadius = ResearchRadius
  MaxChangeAltitude = MaxChangeAltitude
End Sub

Private Sub Class_Initialize()
 ReDim Node(0)
 MaxChangeAltitude = 50
 ResearchRadius = 60
End Sub


