VERSION 1.0 CLASS
BEGIN
  MultiUse = -1  'True
  Persistable = 0  'NotPersistable
  DataBindingBehavior = 0  'vbNone
  DataSourceBehavior  = 0  'vbNone
  MTSTransactionMode  = 0  'NotAnMTSObject
END
Attribute VB_Name = "ParticleSystem8"
Attribute VB_GlobalNameSpace = False
Attribute VB_Creatable = True
Attribute VB_PredeclaredId = False
Attribute VB_Exposed = True
'##CD This class allows you to create great particles effects (such as rain, snow, fountain, cloud, smoke, trails). Usually it is created from the SpecialEffects class but you can create it directly in your program with Public [...] As New BillboardSystem. _
For further information, see <a href="Tutorials/billsys.htm">Billboard system tutorials</a>



' TrueVision 8 3D v0.3
'------------------
'
' Billboard Particles system (for rain, snow, fountain,waterfall, cloud, smoke, trails...)
' ~ by Sylvain Dupont ~
'

' 11-25-2000 : Ported to TV8 with points sprites : faster (1000 particles for 100 FPS !)
' 02-07-2000 : Bouncing & public particle desc
' 30-06-2000 : Vertex Buffer added

Option Explicit
Private BS As BSType
Public Event NewParticle(VelocX, VelocY, VelocZ, ColorR, ColorG, ColorB, ColorA, LifeTime, TextureIndex)
'##ED This event fires when a new particle is created. It's useful to set the new particle properties. The Arguments are passed as references so you have to modity directly these arguments. If you don't change the arguments, the default parameters will be used.
'##PD VelocX Velocity X-coordinate
'##PD VelocY Velocity Y-coordinate
'##PD VelocZ Velocity Z-coordinate
'##PD ColorR New color red value
'##PD ColorG New color green value
'##PD ColorB New color blue value
'##PD ColorA New color alpha value.
'##PD LifeTime New lifetime for this particle (in seconds)
'##PD TextureIndex Texture that will be applied on the particle (Texture index can be obtained with the TextureIndex property of the Texture class)


Private Type ParticleFormat
  Position As D3DVECTOR
  color As Long
  tu As Single
  tv As Single
End Type

Private Const TV_CUSTOMPARTICLEFORMAT = (D3DFVF_XYZ Or D3DFVF_DIFFUSE Or D3DFVF_TEX1)
Private ParticleVB As Direct3DVertexBuffer8

Private Type BillboardParticle
 Texture As Integer
 Position As D3DVECTOR
 Direction As D3DVECTOR
 color As D3DCOLORVALUE
 LifeTime As Single
 Key As Single
 Active As Boolean
End Type
Private Type ParticleAutoGenerate
  Enabled As Boolean
  Veloc As D3DVECTOR
  Random As D3DVECTOR
  color As D3DCOLORVALUE
  LifeTime As Single
End Type
Private Type BSType
  Tri() As D3DVERTEX
  Particle() As BillboardParticle
  Gravity As Single
  BounceFactor As Single
  ParticleCount As Long
  ParticleNow As Long
  GeneratorSpeed As Single
  Speed As Single
  Position As D3DVECTOR
  Radius As Single
  EnergyDropOff As Single
  SphereCenter As D3DVECTOR
  SphereRadius As Single
  TargetPosition As D3DVECTOR
  AutoGenerate As ParticleAutoGenerate
End Type

Private ComputeSphere As Boolean
Private SrcBlend, DestBlend, ChangeMode As Integer
Private CustomPlane As D3DVECTOR, CustomPlaneDist As Single

Public Enum CONST_TV_PARTICLECHANGE
  '##EN Defines the particle color behaviour during the lifetime.
  TV_CHANGE_ALPHA = 1
  '##EM TV_CHANGE_ALPHA Alpha change
  TV_CHANGE_COLOR = 2
  '##EM TV_CHANGE_COLOR Color change
  TV_CHANGE_NO = 3
  '##EM TV_CHANGE_NO No change.
End Enum

Public Enum CONST_TV_PARTICLEUPDATE
 '##EN Defines the update mode of the particle generator. <b><i>Note :</b> Several consts can be added<i>
 TV_UPDATE_ACTIVE = 0
 '##EM TV_UPDATE_ACTIVE Update active particles (that are visible)
 TV_UPDATE_NON_ACTIVE = 1
 '##EM TV_UPDATE_NON_ACTIVE Update non active particles (that aren't visible)
 TV_UPDATE_BOTH = 2
 '##EM TV_UPDATE_BOTH Update all the particles.
 TV_UPDATE_GENERATOR = 4
 '##EM TV_UPDATE_GENERATOR Can generate new particles (Recommanded - With NewParticle event)
 TV_UPDATE_RANDOM_GENERATOR = 8
 '##EM TV_UPDATE_RANDOM_GENERATOR Generate random particles (without NewParticle event)
 TV_UPDATE_FOLLOW_LANDSCAPE = 16
 '##EM TV_UPDATE_FOLLOW_LANDSCAPE The particles have to keep over the landscape (useful for water effect)
 TV_UPDATE_BOUNCE_ON_LANDSCAPE = 32
 '##EM TV_UPDATE_BOUNCE_ON_LANDSCAPE The particles have to bounce on the landscape.
 TV_UPDATE_BOUNCE_ON_Y0_PLANE = 64
 '##EM TV_UPDATE_BOUNCE_ON_Y0_PLANE The particles have to bounce on the plane y=0
 TV_UPDATE_BOUNCE_ON_CUSTOM_PLANE = 128
 '##EM TV_UPDATE_BOUNCE_ON_CUSTOM_PLANE The particles have to bounce on the custom plane defined by the SetBounceCustomPlane method.
End Enum

Public Enum CONST_TV_PARTICLE_PRESET
  TV_PARTICLE_SNOW = 1
  TV_PARTICLE_RAIN = 2
  TV_PARTICLE_EXPLOSION = 3
  TV_PARTICLE_SPARK = 4
  TV_PARTICLE_SMOKE = 5
End Enum
Friend Function LoadFromFile(NumFile) As Boolean
  Dim magic As String * 4
  Get #NumFile, , magic
  If magic = "BSYS" Then
   Get #NumFile, , BS
  End If
End Function
Sub MoveEmettor(NewPos As D3DVECTOR)
  '##BD Moves the generator to the specified position in the 3d world.
  BS.TargetPosition = NewPos
End Sub

Sub RenderParticles()
    Dim v As ParticleFormat, i As Long, PartCount As Long
    Dim TempColor As D3DCOLORVALUE, alpha As Single
    ReDim temppart(BS.ParticleCount) As ParticleFormat
       Device.SetTransform D3DTS_WORLD, WorldMatrix
       With Device
    
        Dim DWFloat0 As Long
        Dim DWFloat1 As Long
        Dim DWFloatp08 As Long
        DWFloat0 = FtoDW(0)
        DWFloat1 = FtoDW(1)
        DWFloatp08 = FtoDW(BS.Radius)
        
        ' Set the render states for using point sprites
        .SetRenderState D3DRS_POINTSPRITE_ENABLE, 1 'True
        .SetRenderState D3DRS_POINTSCALE_ENABLE, 1 'True
        .SetRenderState D3DRS_POINTSIZE, DWFloatp08
        '.SetRenderState
        .SetRenderState D3DRS_POINTSIZE_MIN, DWFloat0
                '***************************************************
        'modified 1/22/01 -Scott Plante- check for max pointsize - 30000 dwoesn't cut it on legacy cards - DO NOT USE 'RESUME NEXT' ON THIS KIND OF STUFF!!!
        .SetRenderState D3DRS_POINT_SIZE_MAX, FtoDW(CSng(TVApp.ParticleSize))
        '***************************************************
        
        .SetRenderState D3DRS_POINTSCALE_A, DWFloat0
        .SetRenderState D3DRS_POINTSCALE_B, DWFloat0
        .SetRenderState D3DRS_POINTSCALE_C, DWFloat1
        .SetRenderState D3DRS_ZWRITEENABLE, 0 'False
        
        ' Set up the vertex buffer to be rendered
        .SetStreamSource 0, ParticleVB, Len(v)
        .SetVertexShader TV_CUSTOMPARTICLEFORMAT
    End With
    PartCount = 0
   For i = 1 To BS.ParticleCount
    With BS.Particle(i)
        If .Active = True Then
            TempColor = .color
            temppart(PartCount).Position = .Position
           If ChangeMode = 1 Then
               alpha = .Key / .LifeTime
               TempColor.a = 1 - alpha
           End If
           If ChangeMode = 2 Then
              alpha = 1 - .Key / .LifeTime
               D3DXColorScale TempColor, .color, alpha
               TempColor.a = .color.a
           End If

            temppart(PartCount).color = DXCToL(TempColor)
            PartCount = PartCount + 1
        End If
    End With
   Next i
   Device.SetRenderState D3DRS_ALPHABLENDENABLE, 1
   Device.SetRenderState D3DRS_DESTBLEND, DestBlend
   Device.SetRenderState D3DRS_SRCBLEND, SrcBlend
   If PartCount > 0 Then
     Device.SetTexture 0, Textures(BS.Particle(1).Texture).TexSurf
     D3DVertexBuffer8SetData ParticleVB, 0, PartCount * Len(v) * 1#, 0, temppart(0)
     Device.DrawPrimitive D3DPT_POINTLIST, 0, PartCount
   End If
   Device.SetRenderState D3DRS_POINTSPRITE_ENABLE, 0
   Device.SetRenderState D3DRS_POINTSCALE_ENABLE, 0
   Device.SetRenderState D3DRS_ZWRITEENABLE, 1 'False

End Sub
'added 1/22/01 -Scott Plante- as part of setting max point size attributes
Public Property Let ParticleSize(vNewValue As Integer)
If vNewValue < 0 Then
    TVApp.ParticleSize = 0
ElseIf vNewValue > TVApp.Caps.MaxPointSize Then
    TVApp.ParticleSize = TVApp.Caps.MaxPointSize
Else
    TVApp.ParticleSize = vNewValue
End If
End Property

Public Property Get ParticleSize() As Integer
  ParticleSize = TVApp.ParticleSize
End Property

Sub ResetAllParticles()
    With BS
      Dim i As Long
      For i = 1 To .ParticleCount
        .Particle(i).Active = False
        .Particle(i).Position = .Position
        .Particle(i).Key = 0
      Next i
    End With
End Sub

Sub SetBounceCustomPlane(nx, ny, nz, Dist)
  '##BD Set the plane on that the particles will bounce. Plane equation : nx*x + ny*y + nz*z + d = 0
  '##PD nx Normal x of the plane
  '##PD ny Normal y of the plane
  '##PD nz Normal z of the plane
  '##PD dist Plane distance
  CustomPlane = Vector(nx, ny, nz)
  CustomPlaneDist = Dist
End Sub

Sub SetGravity(Gravity, Optional BounceFactor As Single = 0.5)
  '##BD Defines the gravity factor and the bounce factor for the particles (less velocity when particles bounce on the floor).
  '##PD Gravity Gravity factor. Usually 1.
  '##PD BounceFactor Bounce factor. Usually 0.5.
  BS.Gravity = Gravity
  BS.BounceFactor = BounceFactor
End Sub


Sub SetParticleSystemPreset(Preset As CONST_TV_PARTICLE_PRESET)
 
End Sub
Sub SetPraticleAutoGenerateMode(Enable As Boolean, Cr As Single, Cg As Single, Cb As Single, Ca As Single, LifeTime As Single, VelocX As Single, VelocY As Single, VelocZ As Single, RandomX As Single, RandomY As Single, RandomZ As Single)
   With BS.AutoGenerate
     .Enabled = Enable
     .color = DXColor(Cr, Cg, Cb, Ca)
     .LifeTime = LifeTime
     .Random = Vector(RandomX, RandomY, RandomZ)
     .Veloc = Vector(VelocX, VelocY, VelocZ)
   End With
End Sub

Friend Sub WriteToFile(NumFile)
 Put #NumFile, , "BSYS"
 Put #NumFile, , BS
End Sub
Public Sub CreateBillboardSystem(ParticleCount As Integer, ParticleStart As Integer, SquareSize As Single, Position As D3DVECTOR, Radius As Single, Texture As Texture8)
  '##BD Create a billboard particles system with the specified settings. Particles can be used for smoke, waterfall, fires, fireworks, fireball, laser sword...
  '##PD ParticleCount Maximum particle count. The engine won't be able to create more particles that this number.
  '##PD ParticleStart Particles already created at the first rendering.
  '##PD SquareSize Face size of the billboard. 2 or 3 for small particles (rain, snow) / 10-50 for big ones (smoke, fire)
  '##PD Position Particle generator position.
  '##PD Radius Not used.
  '##PD Texture Texture will be applied on the billboard.
   On Error GoTo ParticleError
   Dim v As ParticleFormat
   With BS
    .ParticleCount = ParticleCount * 1#
    ReDim .Particle(.ParticleCount)
    .Speed = 0.05

    Dim rad2 As Single
    Dim i As Long
    .Position = Position
    .Radius = SquareSize
    .GeneratorSpeed = 50
    .BounceFactor = 0.5
    .ParticleNow = ParticleStart
    rad2 = Radius / 2
    Randomize Timer
    .TargetPosition = .Position
    Dim PartMem As Long
    PartMem = 1# * Len(v) * ParticleCount
    Set ParticleVB = Device.CreateVertexBuffer(PartMem, 0, TV_CUSTOMPARTICLEFORMAT, D3DPOOL_MANAGED)
    For i = 0 To .ParticleCount
      If i <= .ParticleNow Then
       .Particle(i).Active = True
      Else
       .Particle(i).Active = False
      End If
     .Particle(i).Key = 0
     .Particle(i).LifeTime = 1
     .Particle(i).Direction = Vector(0, Rnd * 3.14, 0)
     '.Particle(i).Position = Vector(Rnd * radius - rad2, Rnd * radius - rad2, Rnd * radius - rad2)
     .Particle(i).Position = .Position
     .Particle(i).Texture = Texture.GetTexIndex
     .Particle(i).color = DXColor(1, 1, 1, 1)
    Next i
     AddLog "PARTICLE : CreateBillboard System :" + CStr(ParticleCount) + " particles"
   End With
   
  
  Exit Sub
ParticleError:
  AddLog "PARTICLE : ERROR WHEN CREATING THE VERTEX BUFFER " + CStr(Err.Description) + " " + CStr(Err.Number)
  Resume Next
End Sub
Function GetActiveParticlesNumber() As Long
 '##BD Returns the particles number that are visible.
  GetActiveParticlesNumber = BS.ParticleNow
End Function

Function GetParticleDirection(Particle As Long) As D3DVECTOR
 '##BD Returns the direction of the specified particle.
  GetParticleDirection = BS.Particle(Particle).Direction
End Function
Function GetParticlePosition(Particle As Long) As D3DVECTOR
   '##BD Returns the position of the specified particle.
   GetParticlePosition = BS.Particle(Particle).Position
End Function

Function GetParticleKey(Particle As Long) As Single
 '##BD Returns the numbers of second since that the specified particle has been created.
 GetParticleKey = BS.Particle(Particle).Key
End Function
Sub SetAlphaBlendingMode(Change As CONST_TV_PARTICLECHANGE, Optional src As CONST_D3DBLEND = D3DBLEND_ONE, Optional dest As CONST_D3DBLEND = D3DBLEND_ONE)
  '##BD Set the alpha blending mode for the particles. This method allows you to decide if the particles will "go out".
  '##PD Change Particles color & alpha changes according to the particle lifetime.
  '##PD src Alpha blending source. Advanced users only.
  '##PD dest Alpha blending destination. Advanced users only.
  DestBlend = dest
  SrcBlend = src
  ChangeMode = Change
End Sub


Sub SetGeneralSystemPosition(Position As D3DVECTOR)
  '##BD Set the particles generator position
  '##PD Position New generator position in the 3d world.
   BS.Position = Position
End Sub
Sub SetGeneratorSpeed(Speed As Single)
 '##BD Defines the speed at which the generator will create new particles.
 '##PD Speed Milliseconds/particle.
 BS.GeneratorSpeed = Speed
End Sub

Sub SetParticleColor(r, g, b, a, Particle)
 '##BD Set the color of the specified particle. <i><b>Note :</b> It's better to use the NewParticle event to set the new particle color, velocity and texture.</i>
 '##PD r New color red-value (0-1)
 '##PD g New color green-value (0-1)
 '##PD b New color blue-value (0-1)
 '##PD a New color alpha-value (0-1)
 '##PD Particle Particle index
 BS.Particle(Particle).color = DXColor(r, g, b, a)
End Sub

Sub SetParticleDirection(Direction As D3DVECTOR, Particle)
 '##BD Set the velocity of the specified particle. <i><b>Note :</b> It's better to use the NewParticle event to set the new particle color, velocity and texture.</i>
 '##PD Direction New velocity vector of the particle.
 '##PD Particle Particle index.
 BS.Particle(Particle).Direction = Direction
End Sub
Sub SetParticleLifeTime(TimeLife As Single, Particle)
 '##BD Set the particle lifetime (in second). <i><b>Note :</b> It's better to use the NewParticle event to set the new particle color, velocity, lifetime and texture.</i>
 '##PD TimeLife New lifetime (seconds)
 '##PD Particle Particle index
 
 BS.Particle(Particle).LifeTime = TimeLife
End Sub

Sub SetParticlePosition(Position As D3DVECTOR, Particle)
  '##BD Set the position of the particle.
  '##PD Position Position New position of the particle.
  '##PD Particle Particle index
  BS.Particle(Particle).Position = Position
End Sub
Public Sub SetParticleSpeed(Speed As Single)
  '##BD Set the general particle speed.
  '##PD Speed Speed of the system.
  
  BS.Speed = Speed
End Sub

Sub SetParticleTexture(Texture As Texture8, Optional Particle As Long = -1)
 '##BD Set the texture for one or all the particles. <i><b>Note :</b> It's better to use the NewParticle event to set the new particle color, velocity and texture.</i>
 '##PD Texture Texture will be applied to the particle
 '##PD Particle Particle index. Nothing or 0 for setting texture for all the particles.
 Dim i As Long

  For i = 1 To BS.ParticleCount
    BS.Particle(i).Texture = Texture.GetTexIndex
  Next i
End Sub


Sub UpdateParticles(Optional UpdateMode As CONST_TV_PARTICLEUPDATE = TV_UPDATE_BOTH)
 '##BD Updates particles with specified mode. This method must be called after the RenderBillboardSystem method.
 '##PD UpdateMode Update type
 Dim Radius As Single, rad2 As Single, textureix As Long
 Dim i As Long
 Dim SpeedFactor As Single, LeftTime As Long, f As Long
 Static AncTime
 On Error GoTo eret

 SpeedFactor = SystemFrame * BS.Speed
 If SystemFrame = 0 Or SystemFrame > 1000 Then Exit Sub
 With BS
   For i = 1 To .ParticleCount
    If .Particle(i).Active = True Then
      .Particle(i).Key = .Particle(i).Key + SystemFrame / 1000
      If .Particle(i).Key > .Particle(i).LifeTime Then
        .Particle(i).Active = False
        .Particle(i).Key = 0
        .ParticleNow = .ParticleNow - 1
      End If
      If UpdateMode And TV_UPDATE_FOLLOW_LANDSCAPE Then
      .Particle(i).Position.x = .Particle(i).Position.x + .Particle(i).Direction.x * SpeedFactor
      .Particle(i).Position.z = .Particle(i).Position.z + .Particle(i).Direction.z * SpeedFactor
      .Particle(i).Position.y = LocalLand.GetHeight(.Particle(i).Position.x + .Position.x, .Particle(i).Position.z + .Position.z)
      Else
       'change particle position      ( + direction )
      .Particle(i).Position.x = .Particle(i).Position.x + .Particle(i).Direction.x * SpeedFactor
      .Particle(i).Position.y = .Particle(i).Position.y + .Particle(i).Direction.y * SpeedFactor
      .Particle(i).Position.z = .Particle(i).Position.z + .Particle(i).Direction.z * SpeedFactor
      .Particle(i).Direction.y = .Particle(i).Direction.y - .Gravity * SystemFrame / 1000
      
      'check bounce :
      If UpdateMode And TV_UPDATE_BOUNCE_ON_LANDSCAPE Then
         If .Particle(i).Position.y < LocalLand.GetHeight(.Particle(i).Position.x, .Particle(i).Position.z) + 5 Then
           .Particle(i).Direction.y = -.Particle(i).Direction.y
           .Particle(i).Position.y = LocalLand.GetHeight(.Particle(i).Position.x, .Particle(i).Position.z) + 5
           D3DXVec3Scale .Particle(i).Direction, .Particle(i).Direction, .BounceFactor
         End If
      End If
      If UpdateMode And TV_UPDATE_BOUNCE_ON_CUSTOM_PLANE Then
       If D3DXVec3Dot(CustomPlane, .Particle(i).Position) - CustomPlaneDist < 0 Then
            'DirectX.VectorReflect .Particle(i).Direction, .Particle(i).Direction, CustomPlane
            .Particle(i).Direction.y = -.Particle(i).Direction.y
           D3DXVec3Scale .Particle(i).Direction, .Particle(i).Direction, .BounceFactor
        End If
      End If
      If UpdateMode And TV_UPDATE_BOUNCE_ON_Y0_PLANE Then
        If .Particle(i).Position.y < 0 Then
           .Particle(i).Position.y = 0
           .Particle(i).Direction.y = -.Particle(i).Direction.y
             D3DXVec3Scale .Particle(i).Direction, .Particle(i).Direction, .BounceFactor
        End If
      End If
      End If
    End If
   Next i
    'Simulate time :
    Dim EmettorSpeed As D3DVECTOR, TempV As D3DVECTOR
    D3DXVec3Subtract EmettorSpeed, .TargetPosition, .Position
    If .GeneratorSpeed < 0.5 Then Exit Sub
    If SystemFrame > 1000 Then SystemFrame = 1000
     LeftTime = SystemFrame + AncTime
     Do
       If LeftTime > .GeneratorSpeed Then
        'find free particle
        f = 0
        For i = 1 To .ParticleCount
          If .Particle(i).Active = False Then f = 1: Exit For
        
        Next i
        If f = 0 Then
          'no more free particles
          GoTo St
        End If
        If SystemFrame > 0 Then
         'the emettor have to continue to move !
         D3DXVec3Scale TempV, EmettorSpeed, (SystemFrame - LeftTime) / SystemFrame
         D3DXVec3Add .Position, .Position, TempV
        End If
        .ParticleNow = .ParticleNow + 1
        .Particle(i).Active = True
        .Particle(i).Key = 0
        .Particle(i).Position = .Position
        'create a new direction (and raise the event):
        Dim Veloc As D3DVECTOR, Col As D3DCOLORVALUE, LifeTime As Single
        
        If .AutoGenerate.Enabled = False Then
         Veloc = Vector(0, -1, 0)
         LifeTime = 1
         Col = DXColor(1, 1, 1, 1)
         textureix = .Particle(i).Texture
         RaiseEvent NewParticle(Veloc.x, Veloc.y, Veloc.z, Col.r, Col.g, Col.b, Col.a, LifeTime, textureix)
         .Particle(i).Direction = Vector3(Veloc.x, Veloc.y, Veloc.z)
         .Particle(i).LifeTime = LifeTime
         .Particle(i).Texture = textureix
         .Particle(i).color = Col
        Else
         .Particle(i).Direction = Vector(.AutoGenerate.Veloc.x + Rnd * .AutoGenerate.Random.x, .AutoGenerate.Veloc.y + Rnd * .AutoGenerate.Random.y, .AutoGenerate.Veloc.z + Rnd * .AutoGenerate.Random.z)
         .Particle(i).LifeTime = .AutoGenerate.LifeTime
         .Particle(i).color = .AutoGenerate.color
        End If
St:
        LeftTime = LeftTime - .GeneratorSpeed
       Else
        .Position = .TargetPosition
        AncTime = LeftTime
        LeftTime = 0
        
       End If
     Loop Until LeftTime <= 0
   End With

 Exit Sub
eret:
 AddLog CStr(Err.Number) + " " + CStr(Err.Description)
 Resume Next
End Sub
Private Sub Class_Initialize()
 SrcBlend = D3DBLEND_ONE
 DestBlend = D3DBLEND_ONE
End Sub

Private Sub Class_Terminate()
  Set ParticleVB = Nothing
  AddLog "PARTICLE SYSTEM : Desallocate memory from vertex buffers"
End Sub


